unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Jpeg, StdCtrls, ExtCtrls, ComCtrls, ShellApi, Math, rkView,
  ImageEnIO, HYIEUtils, HYIEDefs;

const
  CM_UpdateView = WM_USER + 2102; // Custom Message...

type
  PCacheItem = ^TCacheItem;
  TCacheItem = record
    Idx: Integer;
    Size: Integer;
    Age: TDateTime;
    Scale: Integer;
    Bmp: TBitmap;
  end;

  PThumbData = ^TThumbData;
  TThumbData = record
    Name: string;
    ThumbWidth: Word;
    ThumbHeight: Word;
    Size: Integer;
    Modified: TDateTime;
    IWidth, IHeight: Word;
    GotThumb: Boolean;
    Image: TObject;
  end;

  ThumbThread = class(TThread)
  private
    { Private declarations }
    ViewLink: TrkView;
    ItemsLink: TList;
  protected
    procedure Execute; override;
  public
    constructor Create(View: TrkView; Items: TList);
  end;

  TfrmMain = class(TForm)
    stbMain: TStatusBar;
    pnlTop: TPanel;
    pnlBottom: TPanel;
    viewMain: TrkView;
    labInfo: TLabel;
    labText: TLabel;
    btnAdd: TButton;
    tbSize: TTrackBar;
    labThumb: TLabel;
    procedure UpdateStatus;
    procedure BiResample(Src, Dest: TBitmap; Sharpen: Boolean);
    procedure OpenDir;
    procedure btnAddClick(Sender: TObject);
    procedure CMUpdateView(var message: TMessage); message CM_UpdateView;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ItemPaintBasic(Canvas: TCanvas; R: TRect; State: TsvItemState);
    procedure GenCellColors;
    function Running: Boolean;
    procedure Start;
    procedure Stop;
    procedure viewMainCellPaint(Sender: TObject; Canvas: TCanvas; Cell: TRect;
      IdxA, Idx: Integer; State: TsvItemState);
    procedure viewMainSelecting(Sender: TObject; Count: Integer);
    procedure tbSizeChange(Sender: TObject);
    procedure viewMainMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
    Items: TList;
    ThumbSizeW, ThumbSizeH: Integer;

    ThumbJPEG: TJpegImage;

    procedure SetThumbSize(Value: Integer; UpdateTrackbar: Boolean);
    function ThumbBmp(idx: Integer): TBitmap;
    procedure GetThumbImage(Sender: TObject; Thumb: PThumbData);
    procedure ClearThumbs;
    procedure ClearThumbsPool;
  protected
    ThumbThr: ThumbThread;
    ThreadDone: Boolean;
  public
    { Public declarations }
    Directory: string;
    // Colors
    cGSelectedStart,
      cGSelectedEnd,
      cGHotStart,
      cGHotEnd,
      cGDisabledStart,
      cGDisabledEnd,
      cGHeaderStart,
      cGHeaderEnd,
      cGHeaderHotStart,
      cGHeaderHotEnd,
      cGHeaderSelStart,
      cGHeaderSelEnd,
      cHot,
      cSelected,
      cDisabled,
      cBackground,
      cLineHighLight: TColor;
    cShadeSelect: TColor;
    cShadeDisabled: TColor;
    CellShade0: TColor;
    CellShade1: TColor;
    CellShade2: TColor;
    CellShade3: TColor;
    CellShade4: TColor;
    CellBkgColor: TColor;
    CellBrdColor: array[Boolean, Boolean] of TColor;
  end;

var
  frmMain: TfrmMain;
  CellJpeg: TJpegImage;
  WI, HI, TW, TH, HSX, vIdx: Integer;
  CellScale: Integer;
  CellStyle: Integer;
  ThumbsPool: TList;
  ThumbsIO: TImageEnIO;
  PoolSize, MaxPool: Integer;
  Filter: String;

implementation

uses
  BrowseForFolderU;

{$R *.dfm}

function BytesToStr(const i64Size: Int64): string;
const
  i64GB = 1024 * 1024 * 1024;
  i64MB = 1024 * 1024;
  i64KB = 1024;
begin
  if i64Size div i64GB > 0 then
    Result := Format('%.1f GB', [i64Size / i64GB])
  else if i64Size div i64MB > 0 then
    Result := Format('%.2f MB', [i64Size / i64MB])
  else if i64Size div i64KB > 0 then
    Result := Format('%.0f kB', [i64Size / i64KB])
  else
    Result := IntToStr(i64Size) + ' byte';
end;

procedure TfrmMain.UpdateStatus;
var
  i: integer;
  n: Int64;
  item: PThumbData;
begin
  n := 0;
  for i := 0 to viewMain.Selection.Count - 1 do
  begin
    item := Items[viewMain.Selection[i]];
    n := n + item.Size;
  end;
  labInfo.Caption := IntToStr(viewMain.Items.Count) + ' items, ' +
    IntToStr(viewMain.Selection.Count) + ' selected (' + BytesToStr(n) + ')';
end;

// http://www.swissdelphicenter.ch/en/showcode.php?id=1698

function ReadMWord(f: TFileStream): Word;
type
  TMotorolaWord = record
    case Byte of
      0: (Value: Word);
      1: (Byte1, Byte2: Byte);
  end;
var
  MW: TMotorolaWord;
begin
  { It would probably be better to just read these two bytes in normally }
  { and then do a small ASM routine to swap them.  But we aren't talking }
  { about reading entire files, so I doubt the performance gain would be }
  { worth the trouble. }
  f.read(MW.Byte2, SizeOf(Byte));
  f.read(MW.Byte1, SizeOf(Byte));
  Result := MW.Value;
end;

procedure GetJPGSize(const sFile: string; var wWidth, wHeight: Integer);
const
  ValidSig: array[0..1] of Byte = ($FF, $D8);
  Parameterless = [$01, $D0, $D1, $D2, $D3, $D4, $D5, $D6, $D7];
var
  Sig: array[0..1] of byte;
  f: TFileStream;
  x: integer;
  Seg: byte;
  Dummy: array[0..15] of byte;
  Len: word;
  ReadLen: LongInt;
begin
  FillChar(Sig, SizeOf(Sig), #0);
  f := TFileStream.Create(sFile, fmOpenRead);
  try
    ReadLen := f.read(Sig[0], SizeOf(Sig));
    for x := Low(Sig) to High(Sig) do
      if Sig[x] <> ValidSig[x] then
        ReadLen := 0;
    if ReadLen > 0 then
    begin
      ReadLen := f.read(Seg, 1);
      while (Seg = $FF) and (ReadLen > 0) do
      begin
        ReadLen := f.read(Seg, 1);
        if Seg <> $FF then
        begin
          if (Seg = $C0) or (Seg = $C1) then
          begin
            ReadLen := f.read(Dummy[0], 3); { don't need these bytes }
            wHeight := ReadMWord(f);
            wWidth := ReadMWord(f);
          end
          else
          begin
            if not (Seg in Parameterless) then
            begin
              Len := ReadMWord(f);
              f.Seek(Len - 2, 1);
              f.read(Seg, 1);
            end
            else
              Seg := $FF; { Fake it to keep looping. }
          end;
        end;
      end;
    end;
  finally
    f.Free;
  end;
end;

function CalcThumbSize(w, h, tw, th: integer): TPoint;
begin
  Result.X := 0;
  Result.Y := 0;
  if (w < tw) and (h < th) then
  begin
    Result.X := w;
    Result.Y := h;
  end
  else if (w = 0) or (h = 0) then
    Exit
  else
  begin
    if w > h then
    begin
      if w < tw then
        tw := w;
      Result.X := tw;
      Result.Y := Trunc(tw * h / w);
      if Result.Y > th then
      begin
        Result.Y := th;
        Result.X := Trunc(th * w / h);
      end;
    end
    else
    begin
      if h < th then
        th := h;
      Result.Y := th;
      Result.X := Trunc(th * w / h);
      if Result.X > tw then
      begin
        Result.X := tw;
        Result.Y := Trunc(tw * h / w);
      end;
    end;
  end;
end;

procedure MakeThumbNail(Src, Dst: TBitmap);
var
  x, y, ix, iy, w, h, dx, dy: Integer;
  x1, x2, x3: integer;
  RowDest, RowSource, RowSourceStart: Integer;
  iRatio: Integer;
  Ratio: Single;
  iRed, iGrn, iBlu: Integer;
  pt: PRGB24;
  iSrc, iDst: Integer;
  lutW, lutH: array of Integer;
begin
  if (Src.Width <= Dst.Width) and (Src.Height <= Dst.Height) then
  begin
    Dst.Assign(Src);
    Exit;
  end;
  w := Dst.Width;
  h := Dst.Height;
  Ratio := 1 / (w / Src.Width);
  SetLength(lutW, w);
  x1 := 0;
  x2 := Trunc(Ratio);
  for x := 0 to w - 1 do
  begin
    lutW[x] := x2 - x1;
    x1 := x2;
    x2 := Trunc((x + 2) * Ratio);
  end;
  Ratio := 1 / (h / Src.Height);
  SetLength(lutH, h);
  x1 := 0;
  x2 := Trunc(Ratio);
  for x := 0 to h - 1 do
  begin
    lutH[x] := x2 - x1;
    x1 := x2;
    x2 := Trunc((x + 2) * Ratio);
  end;
  RowDest := Integer(Dst.Scanline[0]);
  RowSourceStart := integer(Src.Scanline[0]);
  RowSource := RowSourceStart;
  iDst := ((w * 24 + 31) and not 31) shr 3;
  iSrc := ((Src.Width * 24 + 31) and not 31) shr 3;
  for y := 0 to h - 1 do
  begin
    dy := lutH[y];
    x1 := 0;
    x3 := 0;
    for x := 0 to w - 1 do
    begin
      dx := lutW[x];
      iRed := 0;
      iGrn := 0;
      iBlu := 0;
      RowSource := RowSourceStart;
      for iy := 1 to dy do
      begin
        pt := PRGB24(RowSource + x1);
        for ix := 1 to dx do
        begin
          iRed := iRed + pt.R;
          iGrn := iGrn + pt.G;
          iBlu := iBlu + pt.B;
          inc(pt);
        end;
        RowSource := RowSource - iSrc;
      end;
      iRatio := $00FFFFFF div (dx * dy);
      pt := PRGB24(RowDest + x3);
      pt.R := (iRed * iRatio) shr 24;
      pt.G := (iGrn * iRatio) shr 24;
      pt.B := (iBlu * iRatio) shr 24;
      x1 := x1 + 3 * dx;
      inc(x3, 3);
    end;
    RowDest := RowDest - iDst;
    RowSourceStart := RowSource;
  end;
end;

procedure TfrmMain.BiResample(Src, Dest: TBitmap; Sharpen: Boolean);
// Fast bilinear resampling procedure found at Swiss Delphi Center + my mods...
type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte;
  end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
var
  x, y, px, py: Integer;
  i, x1, x2, z, z2, iz2: Integer;
  w1, w2, w3, w4: Integer;
  Ratio: Integer;
  sDst, sDstOff: Integer;
  sScanLine: array[0..255] of PRGBArray;
  Src1, Src2: PRGBArray;
  C, C1, C2: TRGB24;
  y1, y2, y3, x3, iRed, iGrn, iBlu: Integer;
  p1, p2, p3, p4, p5: PRGB24;
begin
  // ScanLine buffer for Source
  sDst := Integer(src.Scanline[0]);
  sDstOff := Integer(src.Scanline[1]) - sDst;
  for i := 0 to src.Height - 1 do
  begin
    sScanLine[i] := PRGBArray(sDst);
    sDst := sDst + sDstOff;
  end;
  // ScanLine for Destiantion
  sDst := Integer(Dest.Scanline[0]);
  y1 := sDst; // only for sharpening...
  sDstOff := Integer(Dest.Scanline[1]) - sDst;
  // Ratio is same for width and height
  Ratio := ((src.Width - 1) shl 15) div Dest.Width;
  py := 0;
  for y := 0 to Dest.Height - 1 do
  begin
    i := py shr 15;
    if i > src.Height - 1 then
      i := src.Height - 1;
    Src1 := sScanLine[i];
    if i < src.Height - 1 then
      Src2 := sScanLine[i + 1]
    else
      Src2 := Src1;
    z2 := py and $7FFF;
    iz2 := $8000 - z2;
    px := 0;
    for x := 0 to Dest.Width - 1 do
    begin
      x1 := px shr 15;
      x2 := x1 + 1;
      C1 := Src1[x1];
      C2 := Src2[x1];
      z := px and $7FFF;
      w2 := (z * iz2) shr 15;
      w1 := iz2 - w2;
      w4 := (z * z2) shr 15;
      w3 := z2 - w4;
      C.R := (C1.R * w1 + Src1[x2].R * w2 + C2.R * w3 + Src2[x2].R * w4) shr 15;
      C.G := (C1.G * w1 + Src1[x2].G * w2 + C2.G * w3 + Src2[x2].G * w4) shr 15;
      C.B := (C1.B * w1 + Src2[x2].B * w2 + C2.B * w3 + Src2[x2].B * w4) shr 15;
      // Set destination pixel
      PRGBArray(sDst)[x] := C;
      inc(px, Ratio);
    end;
    sDst := sDst + sDstOff;
    inc(py, Ratio);
  end;

  Exit; // Remove this to enable sharpening

  // Sharpening...
  y2 := y1 + sDstOff;
  y3 := y2 + sDstOff;
  for y := 1 to Dest.Height - 2 do
  begin
    for x := 0 to Dest.Width - 3 do
    begin
      x1 := x * 3;
      x2 := x1 + 3;
      x3 := x1 + 6;
      p1 := PRGB24(y1 + x1);
      p2 := PRGB24(y1 + x3);
      p3 := PRGB24(y2 + x2);
      p4 := PRGB24(y3 + x1);
      p5 := PRGB24(y3 + x3);
      // -15 -11                       // -17 - 13
      iRed := (p1.R + p2.R + (p3.R * -15) + p4.R + p5.R) div -11;
      iGrn := (p1.G + p2.G + (p3.G * -15) + p4.G + p5.G) div -11;
      iBlu := (p1.B + p2.B + (p3.B * -15) + p4.B + p5.B) div -11;
      if iRed < 0 then
        iRed := 0
      else if iRed > 255 then
        iRed := 255;
      if iGrn < 0 then
        iGrn := 0
      else if iGrn > 255 then
        iGrn := 255;
      if iBlu < 0 then
        iBlu := 0
      else if iBlu > 255 then
        iBlu := 255;
      PRGB24(y2 + x2).R := iRed;
      PRGB24(y2 + x2).G := iGrn;
      PRGB24(y2 + x2).B := iBlu;
    end;
    inc(y1, sDstOff);
    inc(y2, sDstOff);
    inc(y3, sDstOff);
  end;
end;

function Blend(Color1, Color2: TColor; A: Byte): TColor;
var
  c1, c2: LongInt;
  r, g, b, v1, v2: byte;
begin
  A := Round(2.55 * A);
  c1 := ColorToRGB(Color1);
  c2 := ColorToRGB(Color2);
  v1 := Byte(c1);
  v2 := Byte(c2);
  r := A * (v1 - v2) shr 8 + v2;
  v1 := Byte(c1 shr 8);
  v2 := Byte(c2 shr 8);
  g := A * (v1 - v2) shr 8 + v2;
  v1 := Byte(c1 shr 16);
  v2 := Byte(c2 shr 16);
  b := A * (v1 - v2) shr 8 + v2;
  Result := (b shl 16) + (g shl 8) + r;
end;

procedure TfrmMain.GenCellColors;
begin
  cHot := $00FDDE99;
  cgHotStart := $00FDF5E6;
  cGHotEnd := $00FDFBF6;
  cSelected := $00FDCE99;
  cGSelectedStart := $00FCEFC4;
  cGSelectedEnd := $00FDF8EF;
  cShadeSelect := $00F8F3EA;
  cDisabled := $00D9D9D9;
  cGDisabledStart := $00EAE9E9;
  cGDisabledEnd := $00FCFBFB;
  cShadeDisabled := $00F6F5F5;
  cGHeaderStart := $00F9F9F9;
  cGHeaderEnd := $00FEFEFE;
  cGHeaderHotStart := $00FFEDBD;
  cGHeaderHotEnd := $00FFF7E3;
  cGHeaderSelStart := $00FCEABA;
  cGHeaderSelEnd := $00FCF4E0;
  cBackground := clWindow;
  cLineHighLight := $00FEFBF6;
  CellBkgColor := clWindow;
  CellBrdColor[False, False] := cDisabled;
  CellBrdColor[False, True] := cDisabled;
  CellBrdColor[True, False] := $00B5B5B5;
  CellBrdColor[True, True] := cSelected;
end;

procedure TfrmMain.GetThumbImage(Sender: TObject; Thumb: PThumbData);
type
  TLine8 = array[0..0] of Byte;
  PLine8 = ^Byte;
var
  FName: string;
  ThumbBmp, bmp: TIEBitmap;
  w, h, Sf, jx, jy, ptrS: integer;
  Ext: string;
  MS: TMemoryStream;
  p: TPoint;
  IsJpeg: Boolean;
  X, Y, s0, a0, incS, incA, Alpha: integer;
  iRGB: TRGB24;
begin
  FName := Directory + Thumb.Name;
  bmp := TIEBitmap.Create;
  ThumbsIO.AttachedIEBitmap := bmp;
  ThumbsIO.Params.JPEG_Scale := ioJPEG_FULLSIZE;
  ThumbsIO.ParamsFromFile(FName);
  Ext := LowerCase(ExtractFileExt(FName));
  IsJpeg := (Ext = '.jpeg') or (Ext = '.jpg');
  if IsJpeg then
  begin
    Thumb^.IWidth := ThumbsIO.Params.JPEG_OriginalWidth;
    Thumb^.IHeight := ThumbsIO.Params.JPEG_OriginalHeight;
    Sf := Trunc(Min(Thumb^.IWidth / 255, Thumb^.IHeight / 255));
    if Sf < 0 then
      Sf := 0;
    case Sf of
      0..1: ThumbsIO.Params.JPEG_Scale := ioJPEG_FULLSIZE;
      2..3: ThumbsIO.Params.JPEG_Scale := ioJPEG_HALF;
      4..7: ThumbsIO.Params.JPEG_Scale := ioJPEG_QUARTER;
    else
      ThumbsIO.Params.JPEG_Scale := ioJPEG_EIGHTH;
    end;
  end;
  ThumbsIO.LoadFromFile(FName);
  if ThumbsIO.Aborting then
  begin
    ThumbsIO.IEBitmap.Allocate(32, 32, ie24RGB);
    for jy := 0 to 31 do
    begin
      ptrS := integer(ThumbsIO.IEBitmap.ScanLine[jy]);
      for jx := 0 to 31 do
      begin
        PRGB24(ptrS + jx * 3).r := 255;
        PRGB24(ptrS + jx * 3).G := 0;
        PRGB24(ptrS + jx * 3).B := 0;
      end;
    end;
    Thumb^.ThumbWidth := 32;
    Thumb^.ThumbHeight := 32;
    Thumb^.IWidth := 0;
    Thumb^.IHeight := 0;
  end
  else if not IsJpeg then
  begin
    Thumb^.IWidth := ThumbsIO.Params.Width;
    Thumb^.IHeight := ThumbsIO.Params.Height;
  end;
  ThumbsIO.IEBitmap.PixelFormat := ie24RGB;
  jx := ThumbsIO.Params.Width;
  jy := ThumbsIO.Params.Height;
  if ThumbsIO.Aborting then
  begin
    jx := 32;
    jy := 32;
  end;
  p := CalcThumbSize(jx, jy, 255, 255);
  w := p.X;
  h := p.Y;
  if w <= 0 then
    w := 1;
  if h <= 0 then
    h := 1;
  ThumbBmp := TIEBItmap.Create;
  ThumbBmp.PixelFormat := ie24RGB;
  ThumbBmp.Width := w;
  ThumbBmp.Height := H;
  if bmp.PixelFormat <> ie24RGB then
    bmp.PixelFormat := ie24RGB;
  if bmp.HasAlphaChannel then
  begin
    s0 := integer(bmp.ScanLine[0]);
    incS := integer(bmp.ScanLine[1]) - s0;
    a0 := integer(bmp.AlphaChannel.ScanLine[0]);
    incA := integer(bmp.AlphaChannel.ScanLine[1]) - a0;
    for Y := 0 to bmp.Height - 1 do
    begin
      for X := 0 to bmp.Width - 1 do
      begin
        Alpha := 255 - PLine8(a0 + X)^;
        iRGB := PRGB24(s0 + X * 3)^;
        iRGB.r := (Alpha * ($FF - iRGB.r)) shr 8 + iRGB.r;
        iRGB.G := (Alpha * ($FF - iRGB.G)) shr 8 + iRGB.G;
        iRGB.B := (Alpha * ($FF - iRGB.B)) shr 8 + iRGB.B;
        PRGB24(s0 + X * 3)^ := iRGB;
      end;
      inc(s0, incS);
      inc(a0, incA);
    end;
  end;
  if (bmp.Width <= ThumbBmp.Width) and (bmp.Height <= ThumbBmp.Height) then
    bmp.CopyRectTo(ThumbBmp, 0, 0, 0, 0, bmp.Width, bmp.Height)
  else
    bmp.StretchRectTo(ThumbBmp, 0, 0, ThumbBmp.Width, ThumbBmp.Height, 0, 0,
      bmp.Width, bmp.Height, rfLanczos3, 255);
  ThumbsIO.IEBitmap.Assign(ThumbBmp);
  Thumb^.ThumbWidth := ThumbsIO.IEBitmap.Width;
  Thumb^.ThumbHeight := ThumbsIO.IEBitmap.Height;
  begin
    MS := TMemoryStream.Create;
    try
      ThumbsIO.Params.JPEG_Scale := ioJPEG_FULLSIZE;
      ThumbsIO.Params.JPEG_Quality := 80;
      ThumbsIO.Params.JPEG_OptimalHuffman := True;
      ThumbsIO.SaveToStreamJpeg(MS);
      Thumb^.Image := MS;
    except
      MS.Free;
      Thumb^.Image := nil;
      raise;
    end;
  end;
  bmp.Free;
  ThumbBmp.Free;
  Thumb^.GotThumb := True;
end;

procedure WinGradient( DC: HDC; ARect: TRect; AColor2, AColor1: TColor );
var
  Vertexs: array [ 0 .. 1 ] of TTriVertex;
  GRect: TGradientRect;
begin
  Vertexs[ 0 ].x := ARect.Left;
  Vertexs[ 0 ].y := ARect.Top;
  Vertexs[ 0 ].Red := ( AColor1 and $000000FF ) shl 8;
  Vertexs[ 0 ].Green := ( AColor1 and $0000FF00 );
  Vertexs[ 0 ].Blue := ( AColor1 and $00FF0000 ) shr 8;
  Vertexs[ 0 ].alpha := 0;
  Vertexs[ 1 ].x := ARect.Right;
  Vertexs[ 1 ].y := ARect.Bottom;
  Vertexs[ 1 ].Red := ( AColor2 and $000000FF ) shl 8;
  Vertexs[ 1 ].Green := ( AColor2 and $0000FF00 );
  Vertexs[ 1 ].Blue := ( AColor2 and $00FF0000 ) shr 8;
  Vertexs[ 1 ].alpha := 0;
  GRect.UpperLeft := 0;
  GRect.LowerRight := 1;
  GradientFill( DC, @Vertexs, 2, @GRect, 1, GRADIENT_FILL_RECT_V );
end;

procedure TfrmMain.ItemPaintBasic(Canvas: TCanvas; R: TRect; State:
  TsvItemState);
var
  C: TColor;
begin
  Canvas.Brush.Style := bsClear;
  if ( State = svSelected ) or ( State = svHot ) then
  begin
    if ( ViewMain.Focused) and ( State = svSelected ) then
    begin
      Canvas.Pen.Color := cSelected;
      WinGradient( Canvas.Handle, R, cGSelectedStart, cGSelectedEnd );
    end
    else if ( State = svHot ) then
    begin
      Canvas.Pen.Color := cHot;
      WinGradient( Canvas.Handle, R, cGHotStart, cGHotEnd );
    end
    else
    begin
      Canvas.Pen.Color := cDisabled;
      WinGradient( Canvas.Handle, R, cGDisabledStart, cGDisabledEnd );
    end;
    Canvas.Rectangle( R );
    if ( ViewMain.Focused ) then
      C := cShadeSelect
    else
      C := cShadeDisabled;
    Canvas.Pen.Color := C;
    Canvas.MoveTo( R.Left + 1, R.Top + 2 );
    Canvas.LineTo( R.Left + 1, R.Bottom - 2 );
    Canvas.LineTo( R.Right - 2, R.Bottom - 2 );
    Canvas.LineTo( R.Right - 2, R.Top + 1 );
    Canvas.Pen.Style := psSolid;
    Canvas.Pixels[ R.Left, R.Top ] := C;
    Canvas.Pixels[ R.Left, R.Bottom - 1 ] := C;
    Canvas.Pixels[ R.Right - 1, R.Top ] := C;
    Canvas.Pixels[ R.Right - 1, R.Bottom - 1 ] := C;
  end;
end;

procedure TfrmMain.SetThumbSize(Value: Integer; UpdateTrackbar: Boolean);
var
  w, h: Integer;
begin
  case Value of
    32..63: CellJpeg.Scale := jsQuarter;
    64..127: CellJpeg.Scale := jsHalf;
    128..255: CellJpeg.Scale := jsFullSize;
  else
    CellJpeg.Scale := jsEighth;
  end;
  w := Value;
  h := Value;
  w := w + 20;
  if CellStyle = 0 then
    h := h + 20
  else
    h := h + 40;
  HSX := (w - 70) shr 1; // Help for calculating rating...
  viewMain.CellWidth := w;
  viewMain.CellHeight := h;
  CellScale := Value;
  if UpdateTrackbar then
  begin
    tbSize.OnChange := nil;
    tbSize.Position := CellScale;
    tbSize.OnChange := tbSizeChange;
  end;
  viewMain.CalcView(False);
  if not UpdateTrackBar then
    viewMain.SetAtTop(-1, vIdx);
end;

procedure TfrmMain.Start;
begin
  if Running then
    Exit;
  ThreadDone := False;
  ThumbThr := ThumbThread.Create(viewMain, Items);
end;

procedure TfrmMain.Stop;
begin
  if ThumbThr <> nil then
  begin
    ThumbThr.Terminate;
    ThumbThr.WaitFor;
    ThumbThr.Free;
    ThumbThr := nil;
  end;
end;

procedure TfrmMain.tbSizeChange(Sender: TObject);
begin
  SetThumbSize(tbSize.Position, False);
end;

procedure BiReSample(src, Dest: TBitmap; Sharpen: Boolean); // RMK
type
  TRGB24 = record
    B, G, R: Byte;
  end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
var
  x, y, px, py: Integer;
  i, x1, x2, z, z2, iz2: Integer;
  w1, w2, w3, w4: Integer;
  Ratio: Integer;
  sDst, sDstOff: Integer;
  sScanLine: array[0..255] of PRGBArray;
  Src1, Src2: PRGBArray;
  C, C1, C2: TRGB24;
  y1, y2, y3, x3, iRed, iGrn, iBlu: Integer;
  p1, p2, p3, p4, p5: PRGB24;
begin
  // ScanLine buffer for Source
  sDst := Integer(src.Scanline[0]);
  sDstOff := Integer(src.Scanline[1]) - sDst;
  for i := 0 to src.Height - 1 do
  begin
    sScanLine[i] := PRGBArray(sDst);
    sDst := sDst + sDstOff;
  end;
  // ScanLine for Destiantion
  sDst := Integer(Dest.Scanline[0]);
  y1 := sDst; // only for sharpening...
  sDstOff := Integer(Dest.Scanline[1]) - sDst;
  // Ratio is same for width and height
  Ratio := ((src.Width - 1) shl 15) div Dest.Width;
  py := 0;
  for y := 0 to Dest.Height - 1 do
  begin
    i := py shr 15;
    if i > src.Height - 1 then
      i := src.Height - 1;
    Src1 := sScanLine[i];
    if i < src.Height - 1 then
      Src2 := sScanLine[i + 1]
    else
      Src2 := Src1;
    z2 := py and $7FFF;
    iz2 := $8000 - z2;
    px := 0;
    for x := 0 to Dest.Width - 1 do
    begin
      x1 := px shr 15;
      x2 := x1 + 1;
      C1 := Src1[x1];
      C2 := Src2[x1];
      z := px and $7FFF;
      w2 := (z * iz2) shr 15;
      w1 := iz2 - w2;
      w4 := (z * z2) shr 15;
      w3 := z2 - w4;
      C.R := (C1.R * w1 + Src1[x2].R * w2 + C2.R * w3 + Src2[x2].R * w4) shr 15;
      C.G := (C1.G * w1 + Src1[x2].G * w2 + C2.G * w3 + Src2[x2].G * w4) shr 15;
      C.B := (C1.B * w1 + Src2[x2].B * w2 + C2.B * w3 + Src2[x2].B * w4) shr 15;
      // Set destination pixel
      PRGBArray(sDst)[x] := C;
      inc(px, Ratio);
    end;
    sDst := sDst + sDstOff;
    inc(py, Ratio);
  end;

  if not Sharpen then
    Exit;

  // Sharpening...
  y2 := y1 + sDstOff;
  y3 := y2 + sDstOff;
  for y := 1 to Dest.Height - 2 do
  begin
    for x := 0 to Dest.Width - 3 do
    begin
      x1 := x * 3;
      x2 := x1 + 3;
      x3 := x1 + 6;
      p1 := PRGB24(y1 + x1);
      p2 := PRGB24(y1 + x3);
      p3 := PRGB24(y2 + x2);
      p4 := PRGB24(y3 + x1);
      p5 := PRGB24(y3 + x3);
      // -15 -11                       // -17 - 13
      iRed := (p1.R + p2.R + (p3.R * -15) + p4.R + p5.R) div -11;
      iGrn := (p1.G + p2.G + (p3.G * -15) + p4.G + p5.G) div -11;
      iBlu := (p1.B + p2.B + (p3.B * -15) + p4.B + p5.B) div -11;
      if iRed < 0 then
        iRed := 0
      else if iRed > 255 then
        iRed := 255;
      if iGrn < 0 then
        iGrn := 0
      else if iGrn > 255 then
        iGrn := 255;
      if iBlu < 0 then
        iBlu := 0
      else if iBlu > 255 then
        iBlu := 255;
      PRGB24(y2 + x2).R := iRed;
      PRGB24(y2 + x2).G := iGrn;
      PRGB24(y2 + x2).B := iBlu;
    end;
    inc(y1, sDstOff);
    inc(y2, sDstOff);
    inc(y3, sDstOff);
  end;
end;

function TfrmMain.Running: Boolean;
begin
  Result := ThumbThr <> nil;
end;

function TfrmMain.ThumbBmp(Idx: Integer): TBitmap;
var
  i, n, sf: Integer;
  p: PCacheItem;
  T: PThumbData;
  Bmp, tmp: TBitmap;
  pt: TPoint;
  Oldest: TDateTime;
begin
  Result := nil;
  // if we have thumbs, see if we can find it...
  if ThumbsPool.Count > 0 then
  begin
    i := ThumbsPool.Count - 1;
    while (i >= 0) and (PCacheItem(ThumbsPool[i]).Idx <> Idx) do
      i := i - 1;
    if i <> -1 then
    begin
      p := ThumbsPool[i];
      if (p.Idx = Idx) then
      begin
        if (p.Scale = CellScale) then
        begin
          p.Age := Now;
          Result := p.Bmp
        end
        else
        begin
          PoolSize := PoolSize - p.Size;
          p.Bmp.Free;
          Dispose(p);
          ThumbsPool.Delete(i);
        end;
      end;
    end;
  end;
  // if we dont have a thumb, make one...
  if Result = nil then
  begin
    T := Items[Idx];
    if T.Image <> nil then
    begin
      TMemoryStream(T.Image).Position := 0;

      sf := Trunc(Min(T.ThumbWidth / CellScale, T.ThumbHeight / CellScale));
      if sf < 0 then
        sf := 0;
      case sf of
        0..1: CellJPEG.Scale := jsFullSize;
        2..3: CellJPEG.Scale := jsHalf;
        4..7: CellJPEG.Scale := jsQuarter;
      else
        CellJPEG.Scale := jsEighth;
      end;
      CellJPEG.LoadFromStream(TMemoryStream(T.Image));

      Bmp := TBitmap.Create;
      Bmp.PixelFormat := pf24bit;
      pt := CalcThumbSize(CellJPEG.Width, CellJPEG.Height, CellScale,
        CellScale);
      if pt.x <> CellJPEG.Width then
      begin
        tmp := TBitmap.Create;
        tmp.PixelFormat := pf24bit;
        tmp.Width := CellJPEG.Width;
        tmp.Height := CellJPEG.Height;
        tmp.Canvas.Draw(0, 0, CellJPEG);
        Bmp.Width := pt.x;
        Bmp.Height := pt.y;
        if (Bmp.Width > 4) and (Bmp.Height > 4) then
          BiReSample(tmp, Bmp, False)
        else
          bmp.Canvas.StretchDraw(Rect(0, 0, pt.X, pt.Y), tmp);
        tmp.Free;
      end
      else
      begin
        Bmp.Width := CellJPEG.Width;
        Bmp.Height := CellJPEG.Height;
        Bmp.Canvas.Draw(0, 0, CellJPEG);
      end;
      New(p);
      p.Idx := Idx;
      p.Size := (Bmp.Width * Bmp.Height) * 3;
      p.Age := Now;
      p.Scale := CellScale;
      p.Bmp := Bmp;
      ThumbsPool.Add(p);
      PoolSize := PoolSize + p.Size;
      Result := p.Bmp;
      // Purge thumbs if needed
      while (PoolSize > MaxPool) and (ThumbsPool.Count > 0) do
      begin
        Oldest := Now;
        n := 0;
        for i := 0 to ThumbsPool.Count - 1 do
        begin
          p := ThumbsPool[i];
          if p.Age <= Oldest then
          begin
            Oldest := p.Age;
            n := i;
          end;
        end;
        Assert(n >= 0);
        p := ThumbsPool[n];
        PoolSize := PoolSize - p.Size;
        p.Bmp.Free;
        Dispose(p);
        ThumbsPool.Delete(n);
      end;
    end;
  end;
end;

procedure TfrmMain.viewMainCellPaint(Sender: TObject; Canvas: TCanvas;
  Cell: TRect; IdxA, Idx: Integer; State: TsvItemState);
var
  x, y: Integer;
  F, S: Boolean;
  R: TRect;
  TW, TH: Integer;
  Txt: string;
  Thumb: PThumbData;
  pt: TPoint;
begin
  Thumb := PThumbData(Items[idx]);
  pt := CalcThumbSize(Thumb.ThumbWidth, Thumb.ThumbHeight, CellScale,
    CellScale);
  TW := pt.X;
  TH := pt.Y;
  ItemPaintBasic(Canvas, Cell, State);
  F := viewMain.Focused;
  S := State = svSelected;
  x := Cell.Left + ((Cell.Right - (Cell.Left + TW)) shr 1);
  y := Cell.Top + ((Cell.Bottom - (Cell.Top + TH)) shr 1);
  y := y - 10;
  if (Thumb.Image <> nil) and (Thumb.GotThumb) then
    Canvas.Draw(x, y, ThumbBmp(idx));
  R.Left := X;
  R.Top := Y;
  R.Right := X + TW;
  R.Bottom := Y + TH;
  Canvas.Pen.Color := CellBrdColor[F, S];
  InflateRect(R, 2, 2);
  Canvas.Rectangle(R);
  Canvas.Pen.Color := clWhite;
  InflateRect(R, -1, -1);
  Canvas.Rectangle(R);
  R := Cell;
  R.Top := R.Bottom - 20;
  txt := Thumb.Name;
  DrawText(Canvas.Handle, PChar(txt), Length(txt), R, DT_END_ELLIPSIS or
    DT_SINGLELINE or DT_NOPREFIX or DT_CENTER or DT_VCENTER);
end;

procedure TfrmMain.viewMainMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  Thumb: PThumbData;
begin
  vIdx := viewMain.ViewIdx;
  i := viewMain.ItemAtXY(Point(X, Y), False);
  if i <> -1 then
  begin
    Thumb := Items[viewMain.Items[i]];
    labThumb.Caption := Thumb.Name;
  end
  else
    labThumb.Caption := '';
end;

procedure TfrmMain.viewMainSelecting(Sender: TObject; Count: Integer);
begin
  UpdateStatus;
end;

procedure TfrmMain.OpenDir;
var
  Entry: PThumbData;
  SR: TSearchRec;
  n: Integer;
  Ext: string;
begin
  Directory := BrowseForFolder('Open image folder for viewing', '', False);
  if Directory <> '' then
  begin
    Stop;
    ClearThumbs;
    ClearThumbsPool;
    viewMain.ViewIdx := -1;
    viewMain.Clear;
    Forms.Application.ProcessMessages;
    if Directory[length(Directory)] <> '\' then
      Directory := Directory + '\';
    if FindFirst(Directory + '*.*', faAnyFile - faDirectory, SR) = 0 then
    begin
      Items.Capacity := 1000;
      repeat
        // check if this is an image file we can show
        Ext := LowerCase(ExtractFileExt(SR.Name));
        if Pos(Ext, Filter) <> 0 then       
        begin
          New(Entry);
          Entry.Name := SR.Name;
          Entry.Size := SR.Size;
          Entry.Modified := FileDateToDateTime(SR.Time);
          Entry.IWidth := 0;
          Entry.IHeight := 0;
          Entry.ThumbWidth := 0;
          Entry.ThumbHeight := 0;
          Entry.GotThumb := False;
          Entry.Image := nil;
          n := Items.Add(Entry);
          if n <> -1 then
            viewMain.Items.Add(n);
        end;
      until FindNext(SR) <> 0;
      FindClose(SR);
      Items.Capacity := Items.Count;
    end;
  end;
  viewMain.CalcView(True);
  SetThumbSize(tbSize.Position, False);
  if ThumbThr = nil then
  begin
    ThreadDone := False;
    ThumbThr := ThumbThread.Create(viewMain, Items);
  end;
  UpdateStatus;
end;

procedure TfrmMain.btnAddClick(Sender: TObject);
begin
  OpenDir;
end;

procedure TFrmMain.ClearThumbs;
var
  i: Integer;
  Item: PThumbData;
begin
  viewMain.Items.Clear;
  for i := Items.Count - 1 downto 0 do
  begin
    Item := Items[i];
    if Assigned(Item) then
      if Item.Size <> 0 then
        Item.Image.Free;
    Dispose(Item);
  end;
  Items.Clear;
end;

procedure TFrmMain.ClearThumbsPool;
var
  i: Integer;
  Thumb: PCacheItem;
begin
  for i := ThumbsPool.Count - 1 downto 0 do
  begin
    Thumb := ThumbsPool[i];
    if Thumb.Bmp <> nil then
      Thumb.Bmp.Free;
    Dispose(Thumb);
  end;
  ThumbsPool.Clear;
  PoolSize := 0;
end;

procedure TfrmMain.CMUpdateView(var message: TMessage);
begin
  viewMain.Invalidate;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  Items := TList.Create;
  ThumbJPEG := TJpegImage.Create;
  ThumbJPEG.CompressionQuality := 80;
  ThumbJPEG.Performance := jpBestSpeed;
  // Max thumbnail size...
  ThumbSizeW := 255;
  ThumbSizeH := 255;
  viewMain.CellWidth := ThumbSizeW + 20;
  viewMain.CellHeight := ThumbSizeH + 40;
  CellJpeg := TJpegImage.Create;
  CellJPEG.Performance := jpBestSpeed;
  GenCellColors;
  CellStyle := -1;
  PoolSize := 0;
  MaxPool := Round(((Screen.Width * Screen.Height) * 3) * 1.5);
  ThumbsIO := TImageEnIO.Create(nil);
  Items := TList.Create;
  ThumbsPool := TList.Create;
  Filter := '.jpg;.jpeg;.png;.tiff;.tif;.bmp;.tga;.ico;.gif;.wmf;.cur;.j2k;.jp2;.jif';
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
var
  i: Integer;
  Item: PThumbData;
begin
  ThumbsIO.Free;
  CellJpeg.Free;
  for i := Items.Count - 1 downto 0 do
  begin
    Item := Items[i];
    if Item.Size <> 0 then
      Item.Image.Free;
    Dispose(Item);
  end;
  ThumbsPool.Free;
  Items.Free;
end;

{ ThumbThread }

constructor ThumbThread.Create(View: TrkView; Items: TList);
begin
  ViewLink := View;
  ItemsLink := Items;
  FreeOnTerminate := False;
  inherited Create(False);
  Priority := tpLower;
end;

procedure ThumbThread.Execute;
var
  Cnt, i: Integer;
  PThumb: PThumbData;
  Old: Integer;
  Update: Boolean;
  InView: Integer;
begin
  inherited;
  if (ViewLink.Items.Count = 0) then
    Exit;
  Cnt := 0;
  Old := ViewLink.ViewIdx;
  repeat
    while (Cnt < ViewLink.Items.Count) and (Terminated = False) do
    begin
      if ViewLink.ViewIdx <> Old then
      begin
        Cnt := ViewLink.ViewIdx - 1;
        if Cnt = -1 then
          Cnt := 0;
        Old := ViewLink.ViewIdx;
      end;
      PThumb := PThumbData(ItemsLink.Items[Cnt]);
      Update := PThumb.GotThumb;
      if (not Update) and (not Terminated) then
      begin
        //FrmMain.ThumbsGetThumbnail(ViewLink, PThumb);
        FrmMain.GetThumbImage(ViewLink, PThumb);
        InView := ViewLink.ViewIdx + (ViewLink.ViewColumns * (ViewLink.ViewRows + 1));
        if (Cnt >= ViewLink.ViewIdx) and (Cnt <= InView) then
          PostMessage(FrmMain.Handle, CM_UpdateView, 0, 0);
      end;
      inc(Cnt);
    end;
    Cnt := 0;
    for i := 0 to ViewLink.Items.Count - 1 do
      if PThumbData(ItemsLink.Items[i]).GotThumb = False then
        inc(Cnt);
  until (Cnt = 0) or (Terminated);
  if not Terminated then
    PostMessage(FrmMain.Handle, CM_UpdateView, 0, 0);
end;

end.

